class UnclosedJSONError(Exception):
    """
    Raised when a JSON string is never closed.
    """
    pass

class NotFound(Exception):
    pass

class Unauthorised(Exception):
    pass

class ServerError(Exception):
    pass



from setuptools import setup

setup(
    name='kgreb',
    version='v0.0.1',
    packages=['kgreb', 'kgreb.vle', 'kgreb.util'],
    url='https://github.com/BigPotatoPizzaHey/kgreb/',
    license='',
    author='BigPotatoPizzaHey',
    author_email='',
    description=''
)



from __future__ import annotations

import requests
from bs4 import BeautifulSoup
from urllib.parse import urlparse, parse_qs

from . import file, user, forum
from ..util import commons



class Session:
    def __init__(self, moodle: str):
        self.cookies = {"MoodleSession": moodle}
        self.headers = commons.headers.copy()

        self._sesskey: str | None = None
        self._file_client_id: str | None = None
        self._file_item_id: str | None = None
        self._user_id: int | None = None
        self._user: user.User | None = None

        self.assert_login()


    # --- Session/auth related methods ---
    @property
    def sesskey(self):
        """Get the sesskey query parameter used in various functions"""
        if self._sesskey is None:
            pfx = "var M = {}; M.yui = {};\nM.pageloadstarttime = new Date();\nM.cfg = "

            response = requests.get("https://vle.kegs.org.uk/", cookies=self.cookies, headers=self.headers)
            soup = BeautifulSoup(response.text, "html.parser")

            self._sesskey = None
            for script in soup.find_all("script"):
                text = script.text
                if "\"sesskey\":" in text:
                    i = text.find(pfx)
                    if i > -1:
                        i += len(pfx) - 1
                        data = commons.consume_json(text, i)

                        if isinstance(data, dict):
                            self._sesskey = data.get("sesskey")

        return self._sesskey

    @property
    def file_client_id(self):
        if self._file_client_id is None:
            response = requests.get("https://vle.kegs.org.uk/user/files.php", cookies=self.cookies,
                                    headers=self.headers)
            soup = BeautifulSoup(response.text, "html.parser")

            for div in soup.find_all("div", {"class": "filemanager w-100 fm-loading"}):
                self._file_client_id = div.attrs["id"].split("filemanager-")[1]

        return self._file_client_id

    @property
    def connected_user(self) -> user.User:
        if not self._user:
            self._user = self.connect_user(self.user_id)

        return self._user

    @property
    def file_item_id(self):
        if self._file_item_id is None:
            response = requests.get("https://vle.kegs.org.uk/user/files.php", cookies=self.cookies,
                                    headers=self.headers)
            soup = BeautifulSoup(response.text, "html.parser")
            self._file_item_id = soup.find("input", {"id": "id_files_filemanager"}).attrs.get("value")

        return self._file_item_id

    @property
    def username(self):
        response = requests.get("https://vle.kegs.org.uk/login/index.php", cookies=self.cookies, headers=self.headers)
        soup = BeautifulSoup(response.text, "html.parser")
        for alert_elem in soup.find_all(attrs={"role": "alert"}):
            alert = alert_elem.text

            username = commons.webscrape_value(alert, "You are already logged in as ",
                                               ", you need to log out before logging in as different user.")
            if username:
                return username
        return None

    @property
    def user_id(self):
        if self._user_id is None:
            response = requests.get("https://vle.kegs.org.uk/", headers=self.headers, cookies=self.cookies)
            soup = BeautifulSoup(response.text, "html.parser")

            url = soup.find("a", {"title": "View profile"}) \
                .attrs["href"]

            parsed = parse_qs(urlparse(url).query)
            self._user_id = int(parsed.get("id")[0])

        return self._user_id

    def assert_login(self):
        assert self.username

    # --- Connecting ---
    def connect_user(self, _id: int) -> user.User:
        ret = user.User(_id, _session=self)
        ret.update_from_id()
        return ret

    def connect_forum(self, _id: int) -> forum.Forum:
        ret = forum.Forum(_id, _session=self)
        ret.update_by_id()
        return ret

    # --- Private Files ---
    def _file_data(self, fp: str) -> dict:
        # Believe or not, KegsNet does actually have some JSON endpoints!
        return requests.post("https://vle.kegs.org.uk/repository/draftfiles_ajax.php",
                             params={"action": "list"},
                             data={
                                 "sesskey": self.sesskey,

                                 "clientid": self.file_client_id,
                                 "itemid": self.file_item_id,
                                 "filepath": fp
                             }, cookies=self.cookies, headers=self.headers).json()

    def files_in_dir(self, fp: str):
        data = self._file_data(fp)["list"]
        files = []
        for file_data in data:
            files.append(file.File.from_json(file_data, self))
        return files

    @property
    def files(self):
        return self.files_in_dir('/')

    def add_file(self, title: str, data: bytes, author: str = '', _license: str = "unknown", fp: str = '/'):
        # Perhaps this method should take in a File object instead of title/data/author etc

        requests.post("https://vle.kegs.org.uk/repository/repository_ajax.php",
                      params={"action": "upload"},
                      data={
                          "sesskey": self.sesskey,
                          "repo_id": 3,  # I'm not sure if it has to be 3

                          "title": title,
                          "author": author,
                          "license": _license,

                          "clientid": self.file_client_id,
                          "itemid": self.file_item_id,
                          "savepath": fp
                      },
                      files={"repo_upload_file": data},
                      cookies=self.cookies, headers=self.headers)

        # Save changes
        self.file_save_changes()

    def file_save_changes(self):
        requests.post("https://vle.kegs.org.uk/user/files.php",
                      data={"returnurl": "https://vle.kegs.org.uk/user/files.php",

                            "sesskey": self.sesskey,
                            "files_filemanager": self.file_item_id,
                            "_qf__user_files_form": 1,
                            "submitbutton": "Save changes"},
                      cookies=self.cookies, headers=self.headers)

    @property
    def file_zip(self):
        """
        Returns bytes of your files as a zip archive
        """
        url = requests.post("https://vle.kegs.org.uk/repository/draftfiles_ajax.php",
                            params={"action": "downloaddir"},
                            data={
                                "sesskey": self.sesskey,
                                "client_id": self.file_client_id,
                                "filepath": '/',
                                "itemid": self.file_item_id
                            },

                            headers=self.headers, cookies=self.cookies).json()["fileurl"]
        return requests.get(url, headers=self.headers, cookies=self.cookies).content


# --- * ---

def login(username: str, password: str):
    session = requests.Session()
    response = session.get("https://vle.kegs.org.uk/login/index.php")

    soup = BeautifulSoup(response.text, "html.parser")
    login_token = soup.find("input", {"name": "logintoken"})["value"]

    session.post("https://vle.kegs.org.uk/login/index.php",
                 data={"logintoken": login_token,
                       "anchor": None,
                       "username": username,
                       "password": password
                       }, headers=commons.headers)

    moodle = session.cookies.get("MoodleSession")

    return Session(moodle)


def login_by_moodle(moodle_cookie: str):
    return Session(moodle_cookie)



import json
from datetime import datetime
from typing import Any, Final

from bs4 import BeautifulSoup

from . import exceptions

DIGITS: Final = tuple("0123456789")

headers = {
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.142 Safari/537.36",
    "referer": "https://kegs.org.uk/",
}


def webscrape_value(raw, text_before, text_after, cls: type = str, i1: int = 1, i2: int = 0) -> str | Any:
    return cls(
        raw.split(text_before)[i1] \
            .split(text_after)[i2]
    )


def webscrape_section(raw: str, before: str | int = '', after: str | int = '', cls: type = str) -> str | Any:
    def _toint(val: str | int) -> int:
        return val if isinstance(val, int) else len(val)

    before, after = _toint(before), _toint(after)

    if after == 0:
        section = raw[before:]
    else:
        section = raw[before:-after]
    return cls(section)


def _read_json_number(string: str) -> float | int:
    ret = ''

    minus = string[0] == '-'
    if minus:
        ret += '-'
        string = string[1:]

    def read_fraction(sub: str):
        sub_ret = ''
        if sub[0] == '.':
            sub_ret += '.'
            sub = sub[1:]
            while sub[0] in DIGITS:
                sub_ret += sub[0]
                sub = sub[1:]

        return sub_ret, sub

    def read_exponent(sub: str):
        sub_ret = ''
        if sub[0].lower() == 'e':
            sub_ret += sub[0]
            sub = sub[1:]

            if sub[0] in "-+":
                sub_ret += sub[0]
                sub = sub[1:]

            if sub[0] not in DIGITS:
                raise exceptions.UnclosedJSONError(f"Invalid exponent {sub}")

            while sub[0] in DIGITS:
                sub_ret += sub[0]
                sub = sub[1:]

        return sub_ret

    if string[0] == '0':
        ret += '0'
        string = string[1:]

    elif string[0] in DIGITS[1:9]:
        while string[0] in DIGITS:
            ret += string[0]
            string = string[1:]

    frac, string = read_fraction(string)
    ret += frac

    ret += read_exponent(string)

    return json.loads(ret)


def consume_json(string: str, i: int = 0) -> str | float | int | dict | list | bool | None:
    """
    Reads a JSON string and stops at the natural end (i.e. when brackets close, or when quotes end, etc.)
    """
    # Named by ChatGPT
    section = ''.join(string[i:])
    if section.startswith("true"):
        return True
    elif section.startswith("false"):
        return False
    elif section.startswith("null"):
        return None
    elif section[0] in "0123456789.-":
        return _read_json_number(section)

    depth = 0
    json_text = ''
    out_string = True

    for char in section:
        json_text += char

        if char == '"':
            if len(json_text) > 1:
                unescaped = json_text[-2] != '\\'
            else:
                unescaped = True
            if unescaped:
                out_string ^= True
                if out_string:
                    depth -= 1
                else:
                    depth += 1

        if out_string:
            if char in "[{":
                depth += 1
            elif char in "}]":
                depth -= 1

        if depth == 0 and json_text.strip():
            return json.loads(json_text.strip())

    raise exceptions.UnclosedJSONError(f"Unclosed JSON string, read {json_text}")


def generate_page_range(limit: int, offset: int, items_per_page: int, starting_page: int = 1) -> tuple[
    range, list[int]]:
    """
    Returns a page range (and first indexes per page) generated from a page range
    :param limit: How many items to reach up to
    :param offset: Starting item idx
    :param items_per_page: Number of items per page
    :param starting_page: Page with index '0' (usually 1 actually)
    :return: tuple of page range and list of starting indexes for each page
    """

    if offset < 0:
        raise ValueError(f"offset {offset!r} < 0")
    if limit < 0:
        raise ValueError(f"limit {limit!r} < 0")

    # There are n items on display per page
    # So the first page you need to view is {page 1 idx} + {offset} // {n}

    # The final item to view is at idx {offset} + {limit} - 1
    # (You have to -1 because the index starts at 0)
    # So the page number for this is 1 + (offset + limit - 1) // n

    # But this is a range so we have to add another 1 for the second argument

    page_range = range(
        starting_page + offset // items_per_page,
        1 + starting_page + (offset + limit - 1) // items_per_page
    )

    return (
        page_range,
        [items_per_page * (i - starting_page) for i in page_range]
    )


def find_links(soup: BeautifulSoup) -> list[str]:
    ret = []
    for elem in soup.find_all("a"):
        if "href" in elem.attrs:
            ret.append(str(elem.attrs["href"]))
    return ret


def to_dformat(date: datetime) -> str:
    """
    Convert a datetime to the format dd-mm-yyyy
    """

    def i2zs(v: int):
        """convert an integer to a fixed 2-digit string"""
        return str(v).zfill(2)

    return f"{i2zs(date.month)}-{i2zs(date.day)}-{date.year}"
